@@ -1,7 +1,7 @@
 /*
  * SPDX-License-Identifier: ISC
  *
- * Copyright (c) 2017-2019 Michael Drake <tlsa@netsurf-browser.org>
+ * Copyright (c) 2017-2021 Michael Drake <tlsa@netsurf-browser.org>
  */
 
 /**
@@ -79,12 +79,12 @@ typedef enum cyaml_type {
 	CYAML_MAPPING,
 	/**
 	 * Value is a bit field.  Values of this type require an array of value
-	 * definitions in the schema entry.  If the bitfield is used to store
+	 * definitions in the schema entry.  If the bit field is used to store
 	 * only single-bit flags, it may be better to use \ref CYAML_FLAGS
 	 * instead.
 	 *
 	 * In the YAML, a \ref CYAML_FLAGS value must be presented as a
-	 * mapping of bitfield entry names to their numerical values.
+	 * mapping of bit field entry names to their numerical values.
 	 */
 	CYAML_BITFIELD,
 	/**
@@ -183,7 +183,10 @@ typedef enum cyaml_flag {
 	 * * For \ref CYAML_ENUM, the value becomes the value of the enum.
 	 *   The numerical value is treated as signed.
 	 * * For \ref CYAML_FLAGS, the values are bitwise ORed together.
-	 *   The numerical values are treated as unsigned,
+	 *   The numerical values are treated as unsigned.
+	 *
+	 * For \ref CYAML_FLOAT types, in strict mode floating point values
+	 * that would cause overflow or underflow are not permitted.
 	 */
 	CYAML_FLAG_STRICT   = (1 << 4),
 	/**
@@ -256,6 +259,63 @@ typedef enum cyaml_flag {
 	 *       \ref cyaml_strval strings.
 	 */
 	CYAML_FLAG_CASE_INSENSITIVE = (1 << 8),
+	/**
+	 * When saving, emit scalar value with plain style (no quotes).
+	 *
+	 * \note This is ignored if the value is non-scaler.
+	 *
+	 * \note In cases where conflicting scalar style flags are set, the
+	 *       the one with the highest precedence is used.  From lowest to
+	 *       highest precedence:
+	 *       \ref CYAML_FLAG_SCALAR_PLAIN,
+	 *       \ref CYAML_FLAG_SCALAR_FOLDED,
+	 *       \ref CYAML_FLAG_SCALAR_LITERAL,
+	 *       \ref CYAML_FLAG_SCALAR_QUOTE_SINGLE,
+	 *       \ref CYAML_FLAG_SCALAR_QUOTE_DOUBLE,
+	 *
+	 * If none of these are set, libyaml's default behaviour is used.
+	 */
+	CYAML_FLAG_SCALAR_PLAIN        = (1 <<  9),
+	/**
+	 * When saving, emit scalar value with folded style:
+	 *
+	 * ```yaml
+	 * string: >
+	 *   This string
+	 *   really has no line breaks!
+	 * ```
+	 *
+	 * See the notes for \ref CYAML_FLAG_SCALAR_PLAIN for applicability
+	 * and precedence.
+	 */
+	CYAML_FLAG_SCALAR_FOLDED       = (1 << 10),
+	/**
+	 * When saving, emit scalar value with literal style:
+	 *
+	 * ```yaml
+	 * string: |
+	 *   This is a
+	 *   multi-line string!
+	 * ```
+	 *
+	 * See the notes for \ref CYAML_FLAG_SCALAR_PLAIN for applicability
+	 * and precedence.
+	 */
+	CYAML_FLAG_SCALAR_LITERAL      = (1 << 11),
+	/**
+	 * When saving, emit scalar value with single quotes (`'`).
+	 *
+	 * See the notes for \ref CYAML_FLAG_SCALAR_PLAIN for applicability
+	 * and precedence.
+	 */
+	CYAML_FLAG_SCALAR_QUOTE_SINGLE = (1 << 12),
+	/**
+	 * When saving, emit scalar value with double quotes (`"`).
+	 *
+	 * See the notes for \ref CYAML_FLAG_SCALAR_PLAIN for applicability
+	 * and precedence.
+	 */
+	CYAML_FLAG_SCALAR_QUOTE_DOUBLE = (1 << 13),
 } cyaml_flag_e;
 
 /**
@@ -275,7 +335,7 @@ typedef struct cyaml_strval {
  */
 typedef struct cyaml_bitdef {
 	const char *name; /**< String representing the value's name. */
-	uint8_t offset;   /**< Bit offset to value in bitfield. */
+	uint8_t offset;   /**< Bit offset to value in bit field. */
 	uint8_t bits;     /**< Maximum bits available for value. */
 } cyaml_bitdef_t;
 
@@ -346,7 +406,7 @@ typedef struct cyaml_schema_value {
 		} mapping;
 		/** \ref CYAML_BITFIELD type-specific schema data. */
 		struct {
-			/** Array of bit definitions for the bitfield. */
+			/** Array of bit definitions for the bit field. */
 			const struct cyaml_bitdef *bitdefs;
 			/** Entry count for bitdefs array. */
 			uint32_t count;
@@ -530,7 +590,7 @@ typedef enum cyaml_err {
 	CYAML_ERR_BAD_MIN_MAX_SCHEMA,    /**< Schema minimum exceeds maximum. */
 	CYAML_ERR_BAD_PARAM_SEQ_COUNT,   /**< Bad seq_count param for schema. */
 	CYAML_ERR_BAD_PARAM_NULL_DATA,   /**< Client gave NULL data argument. */
-	CYAML_ERR_BAD_BITVAL_IN_SCHEMA,  /**< Bit value beyond bitfield size. */
+	CYAML_ERR_BAD_BITVAL_IN_SCHEMA,  /**< Bit value beyond bit field size. */
 	CYAML_ERR_SEQUENCE_ENTRIES_MIN,  /**< Too few sequence entries. */
 	CYAML_ERR_SEQUENCE_ENTRIES_MAX,  /**< Too many sequence entries. */
 	CYAML_ERR_SEQUENCE_FIXED_COUNT,  /**< Mismatch between min and max. */
@@ -559,7 +619,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_INT( \
 		_flags, _type) \
 	.type = CYAML_INT, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type)
 
 /**
@@ -576,11 +636,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_INT(((_flags) & (~CYAML_FLAG_POINTER)), \
 				(((_structure *)NULL)->_member)), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -597,11 +657,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_INT(((_flags) | CYAML_FLAG_POINTER), \
 				(*(((_structure *)NULL)->_member))), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -613,7 +673,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_UINT( \
 		_flags, _type) \
 	.type = CYAML_UINT, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type)
 
 /**
@@ -630,11 +690,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_UINT(((_flags) & (~CYAML_FLAG_POINTER)), \
 				(((_structure *)NULL)->_member)), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -651,11 +711,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_UINT(((_flags) | CYAML_FLAG_POINTER), \
 				(*(((_structure *)NULL)->_member))), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -667,7 +727,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_BOOL( \
 		_flags, _type) \
 	.type = CYAML_BOOL, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type)
 
 /**
@@ -684,11 +744,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_BOOL(((_flags) & (~CYAML_FLAG_POINTER)), \
 				(((_structure *)NULL)->_member)), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -705,11 +765,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_BOOL(((_flags) | CYAML_FLAG_POINTER), \
 				(*(((_structure *)NULL)->_member))), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -723,7 +783,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_ENUM( \
 		_flags, _type, _strings, _strings_count) \
 	.type = CYAML_ENUM, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type), \
 	.enumeration = { \
 		.strings = _strings, \
@@ -746,12 +806,12 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _strings, _strings_count) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_ENUM(((_flags) & (~CYAML_FLAG_POINTER)), \
 				(((_structure *)NULL)->_member), \
 				_strings, _strings_count), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -770,12 +830,12 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _strings, _strings_count) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_ENUM(((_flags) | CYAML_FLAG_POINTER), \
 				(*(((_structure *)NULL)->_member)), \
 				_strings, _strings_count), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -789,7 +849,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_FLAGS( \
 		_flags, _type, _strings, _strings_count) \
 	.type = CYAML_FLAGS, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type), \
 	.enumeration = { \
 		.strings = _strings, \
@@ -812,12 +872,12 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _strings, _strings_count) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_FLAGS(((_flags) & (~CYAML_FLAG_POINTER)), \
 				(((_structure *)NULL)->_member), \
 				_strings, _strings_count), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -836,12 +896,12 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _strings, _strings_count) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_FLAGS(((_flags) | CYAML_FLAG_POINTER), \
 				(*(((_structure *)NULL)->_member)), \
 				_strings, _strings_count), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -849,13 +909,13 @@ typedef enum cyaml_err {
  *
  * \param[in]  _flags         Any behavioural flags relevant to this value.
  * \param[in]  _type          The C type for this value.
- * \param[in]  _bitvals       Array of bitfield value data for the bitfield.
+ * \param[in]  _bitvals       Array of bit field value data for the bit field.
  * \param[in]  _bitvals_count Number of entries in _bitvals.
  */
 #define CYAML_VALUE_BITFIELD( \
 		_flags, _type, _bitvals, _bitvals_count) \
 	.type = CYAML_BITFIELD, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type), \
 	.bitfield = { \
 		.bitdefs = _bitvals, \
@@ -865,49 +925,49 @@ typedef enum cyaml_err {
 /**
  * Mapping schema helper macro for keys with \ref CYAML_BITFIELD type.
  *
- * Use this for bitfield types contained in structs.
+ * Use this for bit field types contained in structs.
  *
  * \param[in]  _key        String defining the YAML mapping key for this value.
  * \param[in]  _flags      Any behavioural flags relevant to this value.
  * \param[in]  _structure  The structure corresponding to the mapping.
  * \param[in]  _member     The member in _structure for this mapping value.
- * \param[in]  _bitvals       Array of bitfield value data for the bitfield.
+ * \param[in]  _bitvals       Array of bit field value data for the bit field.
  * \param[in]  _bitvals_count Number of entries in _bitvals.
  */
 #define CYAML_FIELD_BITFIELD( \
 		_key, _flags, _structure, _member, _bitvals, _bitvals_count) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_BITFIELD(((_flags) & (~CYAML_FLAG_POINTER)), \
 				(((_structure *)NULL)->_member), \
 				_bitvals, _bitvals_count), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
  * Mapping schema helper macro for keys with \ref CYAML_BITFIELD type.
  *
- * Use this for pointers to bitfied types.
+ * Use this for pointers to bit field types.
  *
  * \param[in]  _key        String defining the YAML mapping key for this value.
  * \param[in]  _flags      Any behavioural flags relevant to this value.
  * \param[in]  _structure  The structure corresponding to the mapping.
  * \param[in]  _member     The member in _structure for this mapping value.
- * \param[in]  _bitvals       Array of bitfield value data for the bitfield.
+ * \param[in]  _bitvals       Array of bit field value data for the bit field.
  * \param[in]  _bitvals_count Number of entries in _bitvals.
  */
 #define CYAML_FIELD_BITFIELD_PTR( \
 		_key, _flags, _structure, _member, _bitvals, _bitvals_count) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_BITFIELD(((_flags) | CYAML_FLAG_POINTER), \
 				(*(((_structure *)NULL)->_member)), \
 				_bitvals, _bitvals_count), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -919,7 +979,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_FLOAT( \
 		_flags, _type) \
 	.type = CYAML_FLOAT, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type)
 
 /**
@@ -936,11 +996,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_FLOAT(((_flags) & (~CYAML_FLAG_POINTER)), \
 				(((_structure *)NULL)->_member)), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -957,11 +1017,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_FLOAT(((_flags) | CYAML_FLAG_POINTER), \
 				(*(((_structure *)NULL)->_member))), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -984,7 +1044,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_STRING( \
 		_flags, _type, _min, _max) \
 	.type = CYAML_STRING, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type), \
 	.string = { \
 		.min = _min, \
@@ -1007,12 +1067,12 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _min) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_STRING(((_flags) & (~CYAML_FLAG_POINTER)), \
 				(((_structure *)NULL)->_member), _min, \
 				sizeof(((_structure *)NULL)->_member) - 1), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -1035,12 +1095,12 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _min, _max) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_STRING(((_flags) | CYAML_FLAG_POINTER), \
 				(((_structure *)NULL)->_member), \
 				_min, _max), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -1053,7 +1113,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_MAPPING( \
 		_flags, _type, _fields) \
 	.type = CYAML_MAPPING, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type), \
 	.mapping = { \
 		.fields = _fields, \
@@ -1074,11 +1134,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _fields) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_MAPPING(((_flags) & (~CYAML_FLAG_POINTER)), \
 				(((_structure *)NULL)->_member), _fields), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -1096,11 +1156,11 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _fields) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_MAPPING(((_flags) | CYAML_FLAG_POINTER), \
 				(*(((_structure *)NULL)->_member)), _fields), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
@@ -1115,7 +1175,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_SEQUENCE( \
 		_flags, _type, _entry, _min, _max) \
 	.type = CYAML_SEQUENCE, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type), \
 	.sequence = { \
 		.entry = _entry, \
@@ -1160,14 +1220,14 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _entry, _min, _max) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
+	.count_offset = offsetof(_structure, _member ## _count), \
+	.count_size = sizeof(((_structure *)NULL)->_member ## _count), \
 	.value = { \
 		CYAML_VALUE_SEQUENCE((_flags), \
 				(*(((_structure *)NULL)->_member)), \
 				_entry, _min, _max), \
 	}, \
-	.data_offset = offsetof(_structure, _member), \
-	.count_size = sizeof(((_structure *)NULL)->_member ## _count), \
-	.count_offset = offsetof(_structure, _member ## _count), \
 }
 
 /**
@@ -1208,14 +1268,14 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _count, _entry, _min, _max) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
+	.count_offset = offsetof(_structure, _count), \
+	.count_size = sizeof(((_structure *)NULL)->_count), \
 	.value = { \
 		CYAML_VALUE_SEQUENCE((_flags), \
 				(*(((_structure *)NULL)->_member)), \
 				_entry, _min, _max), \
 	}, \
-	.data_offset = offsetof(_structure, _member), \
-	.count_size = sizeof(((_structure *)NULL)->_count), \
-	.count_offset = offsetof(_structure, _count), \
 }
 
 /**
@@ -1229,7 +1289,7 @@ typedef enum cyaml_err {
 #define CYAML_VALUE_SEQUENCE_FIXED( \
 		_flags, _type, _entry, _count) \
 	.type = CYAML_SEQUENCE_FIXED, \
-	.flags = (_flags), \
+	.flags = (enum cyaml_flag)(_flags), \
 	.data_size = sizeof(_type), \
 	.sequence = { \
 		.entry = _entry, \
@@ -1251,12 +1311,12 @@ typedef enum cyaml_err {
 		_key, _flags, _structure, _member, _entry, _count) \
 { \
 	.key = _key, \
+	.data_offset = offsetof(_structure, _member), \
 	.value = { \
 		CYAML_VALUE_SEQUENCE_FIXED((_flags), \
 				(*(((_structure *)NULL)->_member)), \
 				_entry, _count), \
 	}, \
-	.data_offset = offsetof(_structure, _member) \
 }
 
 /**
