@@ -1,7 +1,7 @@
 /*
  * SPDX-License-Identifier: ISC
  *
- * Copyright (C) 2017-2020 Michael Drake <tlsa@netsurf-browser.org>
+ * Copyright (C) 2017-2021 Michael Drake <tlsa@netsurf-browser.org>
  */
 
 /**
@@ -18,6 +18,8 @@
 #include <assert.h>
 #include <limits.h>
 #include <errno.h>
+#include <float.h>
+#include <math.h>
 
 #include <yaml.h>
 
@@ -52,6 +54,10 @@ typedef enum cyaml_event {
 typedef struct cyaml_state {
 	/** Current load state machine state. */
 	enum cyaml_state_e state;
+
+	size_t line;   /**< Event start position in YAML source. */
+	size_t column; /**< Event start position in YAML source. */
+
 	/** Schema for the expected value in this state. */
 	const cyaml_schema_value_t *schema;
 	/** Anonymous union for schema type specific state. */
@@ -639,7 +645,6 @@ static void cyaml__handle_replay(
  * object.
  *
  * \param[in]  ctx    The CYAML loading context.
- * \return \ref CYAML_OK on success, or appropriate error otherwise.
  */
 static void cyaml__delete_yaml_event(
 		cyaml_ctx_t *ctx)
@@ -879,18 +884,21 @@ static cyaml_err_t cyaml__mapping_bitfieid_create(
 		cyaml_ctx_t *ctx,
 		cyaml_state_t *state)
 {
-	cyaml_bitfield_t *bitfield;
 	unsigned count = state->mapping.fields_count;
-	size_t size = ((count + CYAML_BITFIELD_BITS - 1) / CYAML_BITFIELD_BITS)
-			* sizeof(*bitfield);
 
-	bitfield = cyaml__alloc(ctx->config, size, true);
-	if (bitfield == NULL) {
-		return CYAML_ERR_OOM;
+	if (count != 0) {
+		cyaml_bitfield_t *bitfield;
+		size_t size = ((count + CYAML_BITFIELD_BITS - 1) /
+				CYAML_BITFIELD_BITS) * sizeof(*bitfield);
+
+		bitfield = cyaml__alloc(ctx->config, size, true);
+		if (bitfield == NULL) {
+			return CYAML_ERR_OOM;
+		}
+
+		state->mapping.fields_set = bitfield;
 	}
 
-	state->mapping.fields_set = bitfield;
-
 	return CYAML_OK;
 }
 
@@ -996,6 +1004,7 @@ static inline bool cyaml__is_sequence(const cyaml_schema_value_t *schema)
  *
  * \param[in]  ctx     The CYAML loading context.
  * \param[in]  state   The CYAML load state we're pushing a stack entry for.
+ * \param[in]  event   The YAML event we're pushing a stack entry for.
  * \param[in]  schema  The CYAML schema for the value expected in state.
  * \param[in]  data    Pointer to where value's data should be written.
  * \return \ref CYAML_OK on success, or appropriate error code otherwise.
@@ -1003,6 +1012,7 @@ static inline bool cyaml__is_sequence(const cyaml_schema_value_t *schema)
 static cyaml_err_t cyaml__stack_push(
 		cyaml_ctx_t *ctx,
 		enum cyaml_state_e state,
+		const yaml_event_t *event,
 		const cyaml_schema_value_t *schema,
 		cyaml_data_t *data)
 {
@@ -1061,6 +1071,11 @@ static cyaml_err_t cyaml__stack_push(
 			"Load: PUSH[%u]: %s\n", ctx->stack_idx,
 			cyaml__state_to_str(state));
 
+	if (event != NULL) {
+		s.line = event->start_mark.line;
+		s.column = event->start_mark.column;
+	}
+
 	ctx->stack[ctx->stack_idx] = s;
 	ctx->state = ctx->stack + ctx->stack_idx;
 	ctx->stack_idx++;
@@ -1249,25 +1264,34 @@ static void cyaml__backtrace(
 	}
 
 	for (uint32_t idx = ctx->stack_idx - 1; idx != 0; idx--) {
-		cyaml_state_t *state = ctx->stack + idx;
+		const cyaml_state_t *state = ctx->stack + idx;
+
 		switch (state->state) {
 		case CYAML_STATE_IN_MAP_KEY: /* Fall through. */
 		case CYAML_STATE_IN_MAP_VALUE:
 			if (state->mapping.fields_idx !=
 					CYAML_FIELDS_IDX_NONE) {
 				cyaml__log(ctx->config, CYAML_LOG_ERROR,
-						"  in mapping field: %s\n",
+						"  in mapping field '%s' "
+						"(line: %zu, column: %zu)\n",
 						state->mapping.fields[
-						state->mapping.fields_idx].key);
+						state->mapping.fields_idx].key,
+						state->line + 1,
+						state->column + 1);
 			} else {
 				cyaml__log(ctx->config, CYAML_LOG_ERROR,
-						"  in mapping:\n");
+						"  in mapping "
+						"(line: %zu, column: %zu)\n",
+						state->line + 1,
+						state->column + 1);
 			}
 			break;
 		case CYAML_STATE_IN_SEQUENCE:
 			cyaml__log(ctx->config, CYAML_LOG_ERROR,
-					"  in sequence entry: %"PRIu32"\n",
-					state->sequence.count);
+					"  in sequence entry '%"PRIu32"' "
+					"(line: %zu, column: %zu)\n",
+					state->sequence.count,
+					state->line + 1, state->column + 1);
 			break;
 		default:
 			/** \todo \ref CYAML_STATE_IN_DOC handling for multi
@@ -1304,7 +1328,7 @@ static cyaml_err_t cyaml__read_int(
 		return CYAML_ERR_INVALID_DATA_SIZE;
 	}
 
-	max = (INT64_MAX >> ((8 - schema->data_size) * 8)) / 2;
+	max = (int64_t)((UINT64_MAX >> ((8 - schema->data_size) * 8)) / 2);
 	min = (-max) - 1;
 
 	errno = 0;
@@ -1312,6 +1336,9 @@ static cyaml_err_t cyaml__read_int(
 
 	if (end == value || errno == ERANGE ||
 	    temp < min || temp > max) {
+		cyaml__log(ctx->config, CYAML_LOG_ERROR,
+				"Load: Invalid INT value: '%s'\n",
+				value);
 		return CYAML_ERR_INVALID_VALUE;
 	}
 
@@ -1321,11 +1348,13 @@ static cyaml_err_t cyaml__read_int(
 /**
  * Helper to read a number into a uint64_t.
  *
+ * \param[in]  ctx    The CYAML loading context.
  * \param[in]  value  String containing scaler value.
  * \param[in]  out    The place to write the value in the output data.
  * \return \ref CYAML_OK on success, or appropriate error code otherwise.
  */
 static inline cyaml_err_t cyaml__read_uint64_t(
+		const cyaml_ctx_t *ctx,
 		const char *value,
 		uint64_t *out)
 {
@@ -1336,6 +1365,9 @@ static inline cyaml_err_t cyaml__read_uint64_t(
 	temp = strtoull(value, &end, 0);
 
 	if (end == value || errno == ERANGE) {
+		cyaml__log(ctx->config, CYAML_LOG_ERROR,
+				"Load: Invalid uint64_t value: '%s'\n",
+				value);
 		return CYAML_ERR_INVALID_VALUE;
 	}
 
@@ -1368,13 +1400,16 @@ static cyaml_err_t cyaml__read_uint(
 		return CYAML_ERR_INVALID_DATA_SIZE;
 	}
 
-	err = cyaml__read_uint64_t(value, &temp);
+	err = cyaml__read_uint64_t(ctx, value, &temp);
 	if (err != CYAML_OK) {
 		return err;
 	}
 
 	max = (~(uint64_t)0) >> ((8 - schema->data_size) * 8);
 	if (temp > max) {
+		cyaml__log(ctx->config, CYAML_LOG_ERROR,
+				"Load: Invalid UINT value: '%s'\n",
+				value);
 		return CYAML_ERR_INVALID_VALUE;
 	}
 
@@ -1398,7 +1433,7 @@ static cyaml_err_t cyaml__read_bool(
 {
 	bool temp = true;
 	static const char * const false_strings[] = {
-		"false", "no", "disable", "0",
+		"false", "no", "off", "disable", "0",
 	};
 
 	CYAML_UNUSED(ctx);
@@ -1440,11 +1475,14 @@ static cyaml_err_t cyaml__read_enum(
 
 	if (schema->flags & CYAML_FLAG_STRICT) {
 		cyaml__log(ctx->config, CYAML_LOG_ERROR,
-				"Load: Invalid enumeration value: %s\n", value);
+				"Load: Invalid ENUM value: %s\n", value);
 		return CYAML_ERR_INVALID_VALUE;
 
 	}
 
+	cyaml__log(ctx->config, CYAML_LOG_DEBUG,
+			"Load: Attempt numerical fallback for ENUM: "
+			"'%s'\n", value);
 	return cyaml__read_int(ctx, schema, value, data);
 }
 
@@ -1474,8 +1512,31 @@ static cyaml_err_t cyaml__read_float_f(
 	errno = 0;
 	temp = strtof(value, &end);
 
-	if (end == value || errno == ERANGE) {
+	if (end == value) {
+		cyaml__log(ctx->config, CYAML_LOG_ERROR,
+				"Load: Invalid FLOAT value: %s\n", value);
 		return CYAML_ERR_INVALID_VALUE;
+
+	} else if (errno == ERANGE) {
+		cyaml_log_t level = CYAML_LOG_ERROR;
+
+		if (!cyaml__flag_check_all(schema->flags, CYAML_FLAG_STRICT)) {
+			level = CYAML_LOG_NOTICE;
+		}
+
+		if (temp == HUGE_VALF || temp == -HUGE_VALF) {
+			cyaml__log(ctx->config, level,
+				"Load: FLOAT overflow: %s\n", value);
+
+		} else {
+			assert(temp < FLT_MIN || temp > FLT_MAX);
+			cyaml__log(ctx->config, level,
+				"Load: FLOAT underflow: %s\n", value);
+		}
+
+		if (cyaml__flag_check_all(schema->flags, CYAML_FLAG_STRICT)) {
+			return CYAML_ERR_INVALID_VALUE;
+		}
 	}
 
 	memcpy(data, &temp, sizeof(temp));
@@ -1509,8 +1570,24 @@ static cyaml_err_t cyaml__read_float_d(
 	errno = 0;
 	temp = strtod(value, &end);
 
-	if (end == value || errno == ERANGE) {
+	if (end == value) {
+		cyaml__log(ctx->config, CYAML_LOG_ERROR,
+				"Load: Invalid FLOAT value: %s\n", value);
 		return CYAML_ERR_INVALID_VALUE;
+
+	} else if (errno == ERANGE) {
+		cyaml_log_t level = CYAML_LOG_ERROR;
+
+		if (!cyaml__flag_check_all(schema->flags, CYAML_FLAG_STRICT)) {
+			level = CYAML_LOG_NOTICE;
+		}
+
+		cyaml__log(ctx->config, level,
+				"Load: FLOAT overflow/overflow: %s\n", value);
+
+		if (cyaml__flag_check_all(schema->flags, CYAML_FLAG_STRICT)) {
+			return CYAML_ERR_INVALID_VALUE;
+		}
 	}
 
 	memcpy(data, &temp, sizeof(temp));
@@ -1591,8 +1668,14 @@ static cyaml_err_t cyaml__read_string(
 	if (schema->string.min > schema->string.max) {
 		return CYAML_ERR_BAD_MIN_MAX_SCHEMA;
 	} else if (str_len < schema->string.min) {
+		cyaml__log(ctx->config, CYAML_LOG_ERROR,
+				"Load: STRING length < %"PRIu32": %s\n",
+				schema->string.min, value);
 		return CYAML_ERR_STRING_LENGTH_MIN;
 	} else if (str_len > schema->string.max) {
+		cyaml__log(ctx->config, CYAML_LOG_ERROR,
+				"Load: STRING length > %"PRIu32": %s\n",
+				schema->string.max, value);
 		return CYAML_ERR_STRING_LENGTH_MAX;
 	}
 
@@ -1818,7 +1901,7 @@ static cyaml_err_t cyaml__set_bitval(
 
 	switch (cyaml__get_event_type(event)) {
 	case CYAML_EVT_SCALAR:
-		err = cyaml__read_uint64_t(
+		err = cyaml__read_uint64_t(ctx,
 				(const char *)event->data.scalar.value, &value);
 		if (err != CYAML_OK) {
 			return err;
@@ -2046,7 +2129,7 @@ static cyaml_err_t cyaml__read_value(
 		err = cyaml__read_flags_value(ctx, schema, data);
 		break;
 	case CYAML_MAPPING:
-		err = cyaml__stack_push(ctx, CYAML_STATE_IN_MAP_KEY,
+		err = cyaml__stack_push(ctx, CYAML_STATE_IN_MAP_KEY, event,
 				schema, data);
 		break;
 	case CYAML_BITFIELD:
@@ -2054,7 +2137,7 @@ static cyaml_err_t cyaml__read_value(
 		break;
 	case CYAML_SEQUENCE: /* Fall through. */
 	case CYAML_SEQUENCE_FIXED:
-		err = cyaml__stack_push(ctx, CYAML_STATE_IN_SEQUENCE,
+		err = cyaml__stack_push(ctx, CYAML_STATE_IN_SEQUENCE, event,
 				schema, data);
 		break;
 	case CYAML_IGNORE:
@@ -2079,7 +2162,7 @@ static cyaml_err_t cyaml__stream_start(
 		const yaml_event_t *event)
 {
 	CYAML_UNUSED(event);
-	return cyaml__stack_push(ctx, CYAML_STATE_IN_STREAM,
+	return cyaml__stack_push(ctx, CYAML_STATE_IN_STREAM, event,
 			ctx->state->schema, ctx->state->data);
 }
 
@@ -2102,7 +2185,7 @@ static cyaml_err_t cyaml__doc_start(
 		return CYAML_OK;
 	}
 	ctx->state->stream.doc_count++;
-	return cyaml__stack_push(ctx, CYAML_STATE_IN_DOC,
+	return cyaml__stack_push(ctx, CYAML_STATE_IN_DOC, event,
 			ctx->state->schema, ctx->state->data);
 }
 
@@ -2206,7 +2289,7 @@ static cyaml_err_t cyaml__map_key(
 		cyaml_event_t cyaml_event;
 		if (!(ctx->config->flags &
 				CYAML_CFG_IGNORE_UNKNOWN_KEYS)) {
-			cyaml__log(ctx->config, CYAML_LOG_DEBUG,
+			cyaml__log(ctx->config, CYAML_LOG_ERROR,
 					"Load: Unexpected key: %s\n", key);
 			return CYAML_ERR_INVALID_KEY;
 		}
@@ -2228,6 +2311,9 @@ static cyaml_err_t cyaml__map_key(
 	/* Toggle mapping sub-state to value */
 	ctx->state->state = CYAML_STATE_IN_MAP_VALUE;
 
+	ctx->state->line = event->start_mark.line;
+	ctx->state->column = event->start_mark.column;
+
 	return err;
 }
 
@@ -2276,6 +2362,9 @@ static cyaml_err_t cyaml__map_value(
 	 * to move. */
 	state->state = CYAML_STATE_IN_MAP_KEY;
 
+	ctx->state->line = event->start_mark.line;
+	ctx->state->column = event->start_mark.column;
+
 	return cyaml__read_value(ctx, &field->value, data, event);
 }
 
@@ -2296,6 +2385,9 @@ static cyaml_err_t cyaml__seq_entry(
 	uint8_t *value_data = state->data;
 	const cyaml_schema_value_t *schema = state->schema;
 
+	ctx->state->line = event->start_mark.line;
+	ctx->state->column = event->start_mark.column;
+
 	if (state->sequence.count + 1 > state->schema->sequence.max) {
 		cyaml__log(ctx->config, CYAML_LOG_ERROR,
 				"Load: Excessive entries (%"PRIu32" max) "
@@ -2507,7 +2599,7 @@ static cyaml_err_t cyaml__load(
 		return err;
 	}
 
-	err = cyaml__stack_push(&ctx, CYAML_STATE_START, schema, &data);
+	err = cyaml__stack_push(&ctx, CYAML_STATE_START, NULL, schema, &data);
 	if (err != CYAML_OK) {
 		goto out;
 	}
