@@ -1,7 +1,7 @@
 /*
  * SPDX-License-Identifier: ISC
  *
- * Copyright (C) 2018 Michael Drake <tlsa@netsurf-browser.org>
+ * Copyright (C) 2018-2020 Michael Drake <tlsa@netsurf-browser.org>
  */
 
 /**
@@ -216,6 +216,35 @@ static inline yaml_mapping_style_t cyaml__get_emit_style_map(
 	return YAML_ANY_MAPPING_STYLE;
 }
 
+/**
+ * Get the style to use for scalar values from value flags.
+ *
+ * \param[in]  ctx     The CYAML saving context.
+ * \return The libyaml scalar style to emit the value with.
+ */
+static inline yaml_scalar_style_t cyaml__get_emit_style_scalar(
+		const cyaml_schema_value_t *schema)
+{
+	/* Consult flags in order of decreasing priority. */
+	if (schema->flags & CYAML_FLAG_SCALAR_QUOTE_DOUBLE) {
+		return YAML_DOUBLE_QUOTED_SCALAR_STYLE;
+
+	} else if (schema->flags & CYAML_FLAG_SCALAR_QUOTE_SINGLE) {
+		return YAML_SINGLE_QUOTED_SCALAR_STYLE;
+
+	} else if (schema->flags & CYAML_FLAG_SCALAR_LITERAL) {
+		return YAML_LITERAL_SCALAR_STYLE;
+
+	} else if (schema->flags & CYAML_FLAG_SCALAR_FOLDED) {
+		return YAML_FOLDED_SCALAR_STYLE;
+
+	} else if (schema->flags & CYAML_FLAG_SCALAR_PLAIN) {
+		return YAML_PLAIN_SCALAR_STYLE;
+	}
+
+	return YAML_ANY_SCALAR_STYLE;
+}
+
 /**
  * Helper to discern whether to emit document delimiting marks.
  *
@@ -488,7 +517,7 @@ static cyaml_err_t cyaml__emit_scalar(
 	int ret;
 	yaml_event_t event;
 
-	if (schema == NULL) {
+	if (schema->type == CYAML_MAPPING) {
 		cyaml__log(ctx->config, CYAML_LOG_INFO, "Save: [%s]\n", value);
 	} else {
 		cyaml__log(ctx->config, CYAML_LOG_INFO,
@@ -499,7 +528,7 @@ static cyaml_err_t cyaml__emit_scalar(
 			(yaml_char_t *)tag,
 			(yaml_char_t *)value,
 			(int)strlen(value),
-			1, 0, YAML_PLAIN_SCALAR_STYLE);
+			1, 1, cyaml__get_emit_style_scalar(schema));
 
 	return cyaml__emit_event_helper(ctx, ret, &event);
 }
@@ -1136,7 +1165,8 @@ static cyaml_err_t cyaml__write_mapping(
 			}
 		}
 
-		err = cyaml__emit_scalar(ctx, NULL, field->key, YAML_STR_TAG);
+		err = cyaml__emit_scalar(ctx, ctx->state->schema,
+				field->key, YAML_STR_TAG);
 		if (err != CYAML_OK) {
 			return err;
 		}
